# Generated by Django 3.2.6 on 2021-08-20 16:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0022_auto_20210814_1828'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContentKeyValueStore',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('code', models.CharField(max_length=20)),
                ('is_active', models.BooleanField(default=True)),
                ('is_locked', models.BooleanField(default=True)),
                ('form_definition', models.JSONField(blank=True, default=dict)),
                ('content', models.JSONField(blank=True, default=dict)),
            ],
        ),
    ]



from django import forms

from saleboxdjango.lib.shipping_options import SaleboxShippingOptions
from saleboxdjango.views.checkout.base import SaleboxCheckoutBaseView


class SaleboxCheckoutShippingMethodForm(forms.Form):
    shipping_method = forms.IntegerField()


class SaleboxCheckoutShippingMethodView(SaleboxCheckoutBaseView):
    shipping_options_class = SaleboxShippingOptions
    checkout_step = "shipping_method"
    form_class = SaleboxCheckoutShippingMethodForm
    template_name = "salebox/checkout/shipping_method.html"

    def get_additional_context_data(self, context):
        context = self._get_shipping_options(context)

        # print(context['shipping_options'])

        combined_price = None

        # use selected option if exists
        if self.sc.get_raw_data()["shipping_method"]["id"] is not None:
            for o in context["shipping_options"]:
                if o["id"] == self.sc.get_raw_data()["shipping_method"]["id"]:
                    combined_price = o["combined_price"]
                    break

        # no option selected, choice the default
        if combined_price is None:
            for o in context["shipping_options"]:
                if o["available"] and o["selected"]:
                    combined_price = o["combined_price"]
                    break

        # no option selected, no default, fallback to first available
        if combined_price is None:
            for o in context["shipping_options"]:
                if o["available"]:
                    combined_price = o["combined_price"]
                    break

        context["combined_shipping_price"] = combined_price
        return context

    def form_valid_pre_redirect(self, form):
        # retrieve selected option
        id = form.cleaned_data["shipping_method"]
        options = self._get_shipping_options()["shipping_options"]
        selected = next(o for o in options if o["id"] == id)

        # store in checkout
        self.sc.set_shipping_method(
            selected["id"],
            selected["code"],
            selected["label"],
            selected["price"],
            selected["meta"],
            self.request,
        )

        return True

    def _get_shipping_options(self, context={}):
        smc = self.shipping_options_class()
        return smc.go(self.request, self.sc.get_raw_data(), context)



# Generated by Django 3.2.6 on 2021-08-07 06:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0018_auto_20201210_1040'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='boolean_10',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_5',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_6',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_7',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_8',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_9',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='string_10',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='string_5',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='string_6',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='string_7',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='string_8',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='string_9',
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_1',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_10',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_2',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_3',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_4',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_5',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_6',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_7',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_8',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='text_9',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='boolean_5',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='int_5',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='attribute',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='attributeitem',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='basketwishlist',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='callbackstore',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='checkoutstore',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='checkoutstoreupdate',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='country',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='countrystate',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='discountgroup',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='discountruleset',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='event',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='lastupdate',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='member',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='membergroup',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='product',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='level',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='lft',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='rght',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productvariantimage',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productvariantrating',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='translation',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='useraddress',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



from saleboxdjango.models import ContentKeyValueStore


def sync_contentkeyvaluestore(data):
    # get fieldname
    fieldnames = [f.name for f in ContentKeyValueStore._meta.get_fields()]

    # retrieve / create object
    o = ContentKeyValueStore.objects.filter(id=data['id']).first()
    if o is None:
        o = ContentKeyValueStore()

    # define fields to exclude from update
    exclude = []

    # save data
    for key in data:
        if key in fieldnames and key not in exclude:
            setattr(o, key, data[key])
    o.save()


import requests
import hashlib
import datetime

from django.conf import settings


def get(path, payload):
    return requests.get(
        settings.SALEBOX["API"]["URL"] + path,
        headers=generate_headers(),
        params=payload,
    )


def generate_headers():
    pos_id = settings.SALEBOX["API"]["KEY"]
    pos_license = settings.SALEBOX["API"]["LICENSE"]
    timestamp = str(round(datetime.datetime.now(datetime.timezone.utc).timestamp()))
    generated = "".join([pos_id, pos_license, timestamp])
    return {
        "salebox-pos-id": pos_id,
        "salebox-timestamp": timestamp,
        "salebox-secret": hashlib.sha256(generated.encode("utf-8")).hexdigest(),
        "salebox-pos-version": "0.0.247",
    }


def post(path, payload):
    return requests.post(
        settings.SALEBOX["API"]["URL"] + path,
        headers=generate_headers(),
        json=payload,
    )



from django.db.models import Avg, Sum
from django.http import JsonResponse

from saleboxdjango.models import Product, ProductVariant, ProductVariantRating
from saleboxdjango.lib.common import get_rating_display


class SaleboxRating:
    def __init__(self, user=None):
        self.user = user
        self.variant = None  # set via the set_variant_from_id method below

    # used by SaleboxRatingAddView / SaleboxRatingRemoveView to provide rating
    # scores for the variant just modified
    def get_data(self, results):
        results = results.split(",")
        o = {}

        if "global_product_rating" in results:
            o["global_product_rating"] = self.get_global_product_rating()

        if "global_variant_rating" in results:
            o["global_variant_rating"] = self.get_global_variant_rating()

        if "user_variant_rating" in results:
            o["user_rating"] = self.get_user_variant_rating()

        return o

    # get the score for a single product (votes from all users)
    def get_global_product_rating(self):
        if self.variant is None:
            raise ValueError("SaleboxRating product variant not set")

        p = Product.objects.get(id=self.variant.product.id)
        return self._return_rating(p.rating_vote_count, p.rating_score)

    # get reviews (ratings with review text) for a single product (votes from all users)
    # TODO: pagination
    def get_global_product_reviews(self, order="-created"):
        if self.variant is None:
            raise ValueError("SaleboxRating product variant not set")

        # get variant IDs
        variant_ids = ProductVariant.objects.filter(
            product=self.variant.product
        ).values_list("id", flat=True)

        # return all reviews for all variants in above list
        return (
            ProductVariantRating.objects.filter(variant=variant_ids)
            .filter(review__inull=False)
            .filter(active_flag=True)
            .filter(review_qc_flag=False)
            .select_related("user")
            .order_by(order)
        )

    # get the score for a single variant (votes from all users)
    def get_global_variant_rating(self):
        if self.variant is None:
            raise ValueError("SaleboxRating product variant not set")

        pv = ProductVariant.objects.get(id=self.variant.id)
        return self._return_rating(pv.rating_vote_count, pv.rating_score)

    # get reviews (ratings with review text) for a single variant (votes from all users)
    # TODO: pagination
    def get_global_variant_reviews(self, order="-created"):
        if self.variant is None:
            raise ValueError("SaleboxRating product variant not set")

        return (
            ProductVariantRating.objects.filter(variant=self.variant.id)
            .filter(active_flag=True)
            .filter(review__isnull=False)
            .filter(review_qc_flag=False)
            .select_related("user")
            .order_by(order)
        )

    # get the score for a single variant (current user's vote only)
    def get_user_variant_rating(self):
        if self.user is not None:
            if self.variant is None:
                raise ValueError("SaleboxRating product variant not set")

            pvr = (
                ProductVariantRating.objects.filter(variant=self.variant)
                .filter(user=self.user)
                .first()
            )

            if pvr is None:
                return self._return_rating(0, 0)
            else:
                return self._return_rating(1, pvr.rating)

        return None

    # add (or update an existing) variant rating
    def rating_add(self, rating, review=None):
        if self.user is not None:
            if self.variant is None:
                raise ValueError("SaleboxRating product variant not set")

            pvs = ProductVariantRating.objects.filter(variant=self.variant).filter(
                user=self.user
            )

            if len(pvs) > 0:
                pvs[0].rating = rating
                pvs[0].review = review
                pvs[0].save()
                for pv in pvs[1:]:
                    pv.delete()
            else:
                pv = ProductVariantRating(
                    user=self.user, variant=self.variant, rating=rating, review=review
                )
                pv.save()

    # remove a variant rating
    def rating_remove(self):
        if self.user is not None:
            if self.variant is None:
                raise ValueError("SaleboxRating product variant not set")

            pvs = ProductVariantRating.objects.filter(variant=self.variant).filter(
                user=self.user
            )

            for pv in pvs:
                pv.delete()

    # fetch the applicable variant
    def set_variant_from_id(self, variant_id):
        self.variant_id = variant_id
        self.variant = ProductVariant.objects.select_related("product").get(
            id=self.variant_id
        )

    # private function to return results as a dict
    def _return_rating(self, count, rating):
        return {"count": count, "rating": get_rating_display(rating, count)}



from django import forms
from .base import SaleboxCheckoutBaseView


class SaleboxCheckoutConfirmationForm(forms.Form):
    pass


class SaleboxCheckoutConfirmationView(SaleboxCheckoutBaseView):
    checkout_step = "confirmation"
    form_class = SaleboxCheckoutConfirmationForm
    template_name = "salebox/checkout/confirmation.html"



from saleboxdjango.forms import SaleboxRatingAddForm
from saleboxdjango.views.rating.base import SaleboxRatingView


class SaleboxRatingAddView(SaleboxRatingView):
    action = "rating-add"
    form = SaleboxRatingAddForm

    def form_valid(self, request):
        try:
            self.sr.set_variant_from_id(self.form.cleaned_data["variant_id"])
            self.sr.rating_add(self.form.cleaned_data["rating"])

            self.status = "success"
            self.results = self.sr.get_data(self.results_csv)
        except:
            self.status = "fail"



from django import forms
from django.shortcuts import redirect

from saleboxdjango.forms import SaleboxAddressAddForm
from saleboxdjango.lib.address import SaleboxAddress
from saleboxdjango.models import UserAddress
from saleboxdjango.views.checkout.base import SaleboxCheckoutBaseView


class SaleboxCheckoutShippingInvoiceAddressForm(forms.Form):
    shipping_address_id = forms.IntegerField()
    invoice_required = forms.BooleanField(required=False)
    invoice_address_id = forms.IntegerField(required=False)


class SaleboxFormNameForm(forms.Form):
    salebox_form_name = forms.ChoiceField(
        choices=(
            ("select_address", "select_address"),
            ("add_shipping", "add_shipping"),
            ("add_invoice", "add_invoice"),
        )
    )


class SaleboxCheckoutShippingInvoiceAddressView(SaleboxCheckoutBaseView):
    initial = {}
    checkout_step = "shipping_invoice_address"
    form_class = SaleboxCheckoutShippingInvoiceAddressForm
    template_name = "salebox/checkout/shipping_invoice_address.html"

    def get(self, request, *args, **kwargs):
        self.shipping_form = SaleboxAddressAddForm(initial=self.initial)
        self.invoice_form = SaleboxAddressAddForm(initial=self.initial)
        return super().get(self, request, *args, **kwargs)

    def get_additional_context_data(self, context):
        sa = SaleboxAddress(self.request.user)

        # shipping form
        context["shipping_form"] = self.shipping_form
        if self.request.user.is_authenticated:
            context["shipping_addresses"] = sa.get(
                selected_id=self.sc.data["shipping_address"]["id"], force_selected=True
            )
        else:
            context["shipping_addresses"] = []
            if self.sc.data["shipping_address"]["address"]:
                context["shipping_addresses"].append(
                    self.sc.data["shipping_address"]["address"]
                )
                context["shipping_addresses"][0]["selected"] = True

        context["shipping_address_extras"] = sa.form_extras(
            country_id=self.shipping_form["country"].value()
        )

        # invoice form
        context["invoice_form"] = self.invoice_form
        if self.request.user.is_authenticated:
            context["invoice_addresses"] = sa.get(
                selected_id=self.sc.data["invoice_address"]["id"],
                non_null_fields=["tax_id"],
                force_selected=True,
            )
        else:
            context["invoice_addresses"] = []
            if self.sc.data["invoice_address"]["address"]:
                context["invoice_addresses"].append(
                    self.sc.data["invoice_address"]["address"]
                )
                context["invoice_addresses"][0]["selected"] = True
        context["invoice_address_extras"] = sa.form_extras(
            country_id=self.invoice_form["country"].value()
        )

        # form control
        context["shipping_address_id"] = self.sc.data["shipping_address"]["id"]
        context["invoice_address_id"] = self.sc.data["invoice_address"]["id"]
        context["invoice_required"] = self.sc.data["invoice_address"]["required"]
        return context

    def post(self, request, *args, **kwargs):
        self.shipping_form = SaleboxAddressAddForm(initial=self.initial)
        self.invoice_form = SaleboxAddressAddForm(initial=self.initial)

        # which action is being performed?
        form = SaleboxFormNameForm(request.POST)
        if form.is_valid():
            action = form.cleaned_data["salebox_form_name"]
        else:
            return self.get(request, *args, **kwargs)

        # action: user has selected an option
        if action == "select_address":
            if request.user.is_authenticated:
                form = self.form_class(request.POST)
                if form.is_valid():
                    self.sc.set_shipping_address(
                        True, form.cleaned_data["shipping_address_id"], None, request
                    )
                    self.sc.set_invoice_address(
                        form.cleaned_data["invoice_required"],
                        form.cleaned_data["invoice_address_id"],
                        None,
                        request,
                    )

                    r = self.sc.set_completed(self.checkout_step, request)
                    if r is None:
                        raise Exception(
                            "There is no next checkout step to redirect to..."
                        )
                    else:
                        return redirect(r)

            # unathenticated users
            if not request.user.is_authenticated and self.sc.data["shipping_address"]:
                r = self.sc.set_completed(self.checkout_step, request)
                if r is None:
                    raise Exception("There is no next checkout step to redirect to...")
                else:
                    return redirect(r)

        # action: user is adding a shipping address
        if action == "add_shipping":
            self.shipping_form = SaleboxAddressAddForm(
                request.POST, initial=self.initial
            )
            if self.shipping_form.is_valid():
                # add address to db
                sa = SaleboxAddress(request.user)
                address = sa.add(self.shipping_form.cleaned_data)
                self.sc.set_shipping_address(True, address, None, request)

                # redirect to self to prevent refresh
                return redirect(request.get_full_path())

        # action: user is adding an invoice address
        if action == "add_invoice":
            self.invoice_form = SaleboxAddressAddForm(
                request.POST, initial=self.initial
            )
            if self.invoice_form.is_valid():
                tax_id = self.invoice_form.cleaned_data["tax_id"]
                if tax_id is not None and len(tax_id.strip()) > 4:
                    # add address to db
                    sa = SaleboxAddress(request.user)
                    address = sa.add(self.invoice_form.cleaned_data)
                    self.sc.set_invoice_address(True, address, None, request)

                    # redirect to self to prevent refresh
                    return redirect(request.get_full_path())

        # if we've reached this point, some input was invalid...
        # just re-show the page
        return self.form_invalid(form)



from django.core.management.base import BaseCommand, CommandError
from rawquery import RawQuery

# run this command immediately AFTER running django's clearsessions


class Command(BaseCommand):
    def handle(self, *args, **options):
        rq = RawQuery()
        MAX_SESSIONS = 5000
        total_sessions_deleted = 0

        while True:
            oldest_sessions = rq.multiple_values(
                f"""
                SELECT      session
                FROM        saleboxdjango_basketwishlist
                WHERE       user_id IS NULL
                GROUP BY    session
                ORDER BY    MIN(last_update)
                LIMIT       {MAX_SESSIONS}
                """
            )

            active_sessions = rq.multiple_values(
                f"""
                SELECT      session_key
                FROM        django_session
                WHERE       session_key IN ({self.make_sql_list(oldest_sessions)})
                """
            )

            sessions_to_delete = [
                s for s in oldest_sessions if s not in active_sessions
            ]

            if len(sessions_to_delete) > 0:
                total_sessions_deleted += len(sessions_to_delete)
                rq.run(
                    f"""
                    DELETE FROM     saleboxdjango_basketwishlist
                    WHERE           user_id IS NULL
                    AND             session IN ({self.make_sql_list(sessions_to_delete)})
                    """
                )
            else:
                break

        print(f"Total sessions deleted: {total_sessions_deleted}")

    def make_sql_list(self, id_list):
        return ",".join(["'" + s + "'" for s in id_list])



from django.core.cache import cache

from saleboxdjango.models import Product, ProductCategory, ProductVariant


class SaleboxCategory:
    def __init__(self):
        self.populated_categories_only = True
        self.product_attribute_include = {}
        self.product_attribute_exclude = {}
        self.variant_attribute_include = {}
        self.variant_attribute_exclude = {}

        self.valid_ids = []

    def get_tree(self, cache_key=None, cache_timeout=86400, category_id=None):
        # attempt to get the complete tree dict from cache
        if cache_key is not None:
            detail_key = "%s::%s" % (cache_key, category_id)
            complete_tree = cache.get(detail_key)
            if complete_tree is not None:
                return complete_tree

        # complete tree does not exist, constuct it
        inner_tree = self._get_tree(cache_key, cache_timeout)
        current = self._find_node(inner_tree, category_id) if category_id else None
        output = {
            "tree": inner_tree,
            "current": current,
            "ancestors": self._find_ancestors(
                inner_tree, [], current["parent"] if current else None
            ),
        }

        # store the complete tree dict in the cache
        if cache_key is not None:
            cache.set(detail_key, output, cache_timeout)

        # return
        return output

    def _find_ancestors(self, tree, ancestors, id):
        if id is not None:
            curr = self._find_node(tree, id)
            del curr["children"]
            if curr is not None:
                ancestors.append(curr)
                if curr["parent"] is not None:
                    ancestors = self._find_ancestors(tree, ancestors, curr["parent"])

        return ancestors

    def _find_node(self, tree, id):
        for c in tree:
            if id == c["id"]:
                return c
            elif len(c["children"]) > 0:
                res = self._find_node(c["children"], id)
                if res is not None:
                    return res

    def _get_root_categories(self):
        tree = []
        categories = ProductCategory.objects.filter(active_flag=True).order_by("name")
        for c in categories:
            if c.is_root_node():
                tree.append(c)

        return tree

    def _get_product_count(self, category):
        ids = category.get_descendants(include_self=True).values_list("id", flat=True)

        pv = (
            ProductVariant.objects.filter(product__category__id__in=list(ids))
            .filter(product__active_flag=True)
            .filter(active_flag=True)
            .filter(available_on_ecom=True)
        )
        if len(self.product_attribute_include.keys()) > 0:
            pv = pv.filter(**self.product_attribute_include)

        if len(self.product_attribute_exclude.keys()) > 0:
            pv = pv.exclude(**self.product_attribute_exclude)

        if len(self.variant_attribute_include.keys()) > 0:
            pv = pv.filter(**self.variant_attribute_include)

        if len(self.variant_attribute_exclude.keys()) > 0:
            pv = pv.exclude(**self.variant_attribute_exclude)

        return pv.order_by("product__id").distinct("product__id").count()

    def _get_tree(self, cache_key, cache_timeout):
        tree = None

        # attempt to retrieve from the cache if a key is defined
        if cache_key is not None:
            tree = cache.get(cache_key)

        # build tree if not in cache
        if tree is None:
            tree = self._get_tree_recurse(self._get_root_categories())
            if cache_key is not None:
                cache.set(cache_key, tree, cache_timeout)

        #
        return tree

    def _get_tree_recurse(self, categories):
        output = []

        for c in categories:
            count = self._get_product_count(c)
            if self.populated_categories_only and count == 0:
                continue

            children = self._get_tree_recurse(c.get_children().filter(active_flag=True))

            output.append(
                {
                    "id": c.id,
                    "children": children,
                    "i18n": c.i18n,
                    "image": c.local_image,
                    "name": c.name,
                    "parent": c.parent.id if c.parent else None,
                    "product_count": count,
                    "short_name": c.short_name,
                    "slug": c.slug,
                    "slug_path": c.slug_path,
                }
            )

        return output

    def set_product_attribute_include(self, attribute_number, value):
        key = "product__attribute_%s" % attribute_number
        self.product_attribute_include[key] = value

    def set_product_attribute_include_keyvalue(
        self, attribute_number, field_name, field_value, field_modifier=None
    ):
        key = "product__attribute_%s__%s" % (attribute_number, field_name)
        if field_modifier is not None:
            key = "%s__%s" % (key, field_modifier)
        self.product_attribute_include[key] = field_value

    def set_product_attribute_exclude(self, attribute_number, value):
        key = "product__attribute_%s" % attribute_number
        self.product_attribute_exclude[key] = value

    def set_product_attribute_exclude_keyvalue(
        self, attribute_number, field_name, field_value, field_modifier=None
    ):
        key = "product__attribute_%s__%s" % (attribute_number, field_name)
        if field_modifier is not None:
            key = "%s__%s" % (key, field_modifier)
        self.product_attribute_exclude[key] = field_value

    def set_variant_attribute_include(self, attribute_number, value):
        key = "attribute_%s" % attribute_number
        self.variant_attribute_include[key] = value

    def set_variant_attribute_include_keyvalue(
        self, attribute_number, field_name, field_value, field_modifier=None
    ):
        key = "attribute_%s__%s" % (attribute_number, field_name)
        if field_modifier is not None:
            key = "%s__%s" % (key, field_modifier)
        self.variant_attribute_include[key] = field_value

    def set_variant_attribute_exclude(self, attribute_number, value):
        key = "attribute_%s" % attribute_number
        self.variant_attribute_exclude[key] = value

    def set_variant_attribute_exclude_keyvalue(
        self, attribute_number, field_name, field_value, field_modifier=None
    ):
        key = "attribute_%s__%s" % (attribute_number, field_name)
        if field_modifier is not None:
            key = "%s__%s" % (key, field_modifier)
        self.variant_attribute_exclude[key] = field_value



# Generated by Django 2.1.8 on 2019-11-13 07:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0011_auto_20191006_1122'),
    ]

    operations = [
        migrations.AddField(
            model_name='productvariant',
            name='string_5',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='string_1',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='string_2',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='string_3',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='string_4',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



# Generated by Django 3.2.12 on 2024-03-15 00:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0028_alter_analytic_options'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='basketwishlist',
            options={'ordering': ['-last_update']},
        ),
    ]



from saleboxdjango.views.basket.base import SaleboxBasketView
from saleboxdjango.forms import SaleboxBasketMigrateForm


class SaleboxBasketMigrateView(SaleboxBasketView):
    action = "basket-migrate"
    form = SaleboxBasketMigrateForm

    def form_valid(self, request):
        try:
            self.sb.migrate_basket_wishlist(
                request,
                self.form.cleaned_data["variant_id"],
                self.form.cleaned_data["to_basket"],
            )
            self.status = "success"

            self.results = self.sb.get_data(request, self.results_csv)
        except:
            self.status = "fail"



from django.db.models import Count

from saleboxdjango.lib.api import get
from saleboxdjango.management.commands.lib.log import log
from saleboxdjango.management.commands.lib.models.contentkeyvaluestore import sync_contentkeyvaluestore
from saleboxdjango.management.commands.lib.models.contentpage import sync_contentpage
from saleboxdjango.management.commands.lib.models.contentpageitem import sync_contentpageitem
from saleboxdjango.management.commands.lib.models.productcategory import sync_productcategory
from saleboxdjango.management.commands.lib.models.product import sync_product
from saleboxdjango.management.commands.lib.models.productvariant import sync_productvariant

from saleboxdjango.models import SyncQueue


class SaleboxPullModels:
    def __init__(self):
        self.error = False

        # retreive summary of queue size
        items_in_queue, config = self.retrieve_queue_size()
        if not items_in_queue:
            return

        # set dependency levels, i.e. items in level 1 depend on level 0, level 2 on level 1, etc
        groups = self.get_dependency_groups()
        for level in groups:
            for model_name in level:
                if model_name in config and config[model_name] > 0 and not self.error:
                    self.sync_models(model_name)

    def get_dependency_groups(self):
        one = [
            'content_contentkeyvaluestore',
            'content_contentpage',
            'product_productcategory',
        ]

        two = [
            'content_contentpageitem',
            'product_product',
        ]

        three = [
            'product_productvariant',
        ]

        return [one, two, three]

    def retrieve_queue_size(self):
        data = SyncQueue \
                .objects \
                .values('model_name') \
                .annotate(count=Count('model_name')) \
                .order_by('model_name') \
                .values_list('model_name', 'count')

        output = {}
        total = 0
        for d in data:
            output[d[0]] = d[1]
            total += d[1]

        return (total > 0, output)

    def sync_models(self, model_name):
        function = {
            'content_contentkeyvaluestore': sync_contentkeyvaluestore,
            'content_contentpage': sync_contentpage,
            'content_contentpageitem': sync_contentpageitem,
            'product_productcategory': sync_productcategory,
            'product_product': sync_product,
            'product_productvariant': sync_productvariant,
        }

        queue_ids = SyncQueue \
                        .objects \
                        .filter(model_name=model_name) \
                        .order_by('created') \
                        .values_list('model_id', flat=True)

        # spit queue into batches of 100, then process each batch in turn
        batch_size = 100
        queue_id_batches = [queue_ids[i * batch_size:(i + 1) * batch_size] for i in range((len(queue_ids) + batch_size - 1) // batch_size )]
        for batch in queue_id_batches:
            path = f"/api/v1/sync/{model_name.replace('_', '-')}"
            r = get(path, {'id': ','.join(batch)})
            if r.status_code != 200:
                log('sync_pull_models', f'API call failed: {path} status: {r.status_code}', 'ERROR')
                self.error = True
                return

            # process objects returned from the api
            response = r.json()
            for o in response['objects']:
                function[response['model']](o)

            # remove all completed items from the queue
            SyncQueue \
                .objects \
                .filter(model_name=model_name) \
                .filter(model_id__in=batch) \
                .delete()



from pprint import pprint

from pyshipping.package import Package
from pyshipping.binpack_simple import binpack


class BinPack:
    def __init__(self):
        self.bins = []
        self.bin_lookup = {}
        self.packages = []
        self.oversize = []
        self.min_bin_size = -1
        self.options = []

    def add_bin(self, id, price, height, width, depth):
        # add bin
        pkg = Package((height, width, depth))
        self.bins.append(
            {
                "id": id,
                "price": price,
                "pkg": pkg,
            }
        )

        # sort bins, cheapest first
        self.bins = sorted(self.bins, key=lambda x: (x["price"]))

        # update bin lookup
        for i, b in enumerate(self.bins):
            self.bin_lookup[b["id"]] = i

    def add_package(self, meta, height, width, depth):
        pkg = Package((height, width, depth))

        # set smallest container size
        oversize = True
        for i, b in enumerate(self.bins):
            good, bad = binpack([pkg], b["pkg"])
            if len(good) == 1:
                oversize = False
                if i > self.min_bin_size:
                    self.min_bin_size = i
                break

        # add to appropriate list
        if oversize:
            self.oversize.append({"meta": meta, "pkg": pkg})
        else:
            self.packages.append(
                {
                    "meta": meta,
                    "min_bin_size": i,
                    "pkg": pkg,
                }
            )

        # sort packages, largest volume first
        self.packages = sorted(
            self.packages, key=lambda x: x["pkg"].volume, reverse=True
        )

    def calculate(self, bin_sizes):
        containers = []

        for p in self.packages:
            containers.append(
                {
                    "contents": {
                        "meta": [],
                        "packages": [],
                    },
                    "id": self.bins[bin_sizes[-1]]["id"],
                }
            )

        # loop through packages
        for p in self.packages:
            for c in containers:
                good, bad = binpack(
                    c["contents"]["packages"] + [p["pkg"]],
                    self.bins[self.bin_lookup[c["id"]]]["pkg"],
                )
                if len(good) == 1:
                    c["contents"]["packages"] += [p["pkg"]]
                    c["contents"]["meta"] += [p["meta"]]
                    break

        # remove empty bins
        for i in reversed(range(len(containers))):
            if len(containers[i]["contents"]["packages"]) == 0:
                containers.pop()

        # attempt to reduce the size of the bins
        for c in containers:
            for b in self.bins:
                good, bad = binpack(c["contents"]["packages"], b["pkg"])
                if len(good) == 1 and len(bad) == 0:
                    c["id"] = b["id"]
                    break

        return containers

    def go(self):
        tmp = []
        results = []

        # calculate bin size combinations
        bin_sizes = list(range(self.min_bin_size, len(self.bins)))
        while len(bin_sizes) > 0:
            tmp.append(self.calculate(bin_sizes))
            bin_sizes.pop()

        # calculate prices
        for t in tmp:
            total_price = 0
            for c in t:
                c["price"] = self.bins[self.bin_lookup[c["id"]]]["price"]
                total_price += c["price"]

            results.append(
                {"price": total_price, "number_of_packages": len(t), "packages": t}
            )

        # remove duplicate options
        duplicates = []
        for r in results:
            s = "%s:%s" % (r["number_of_packages"], r["price"])
            if s not in duplicates:
                duplicates.append(s)
            else:
                r["duplicate"] = True
        results = [r for r in results if "duplicate" not in r]

        # sort prices, cheapest first
        results = sorted(
            results,
            key=lambda x: (x["price"], x["number_of_packages"]),
        )

        return {"results": results, "oversize": [o["meta"] for o in self.oversize]}

    def debug(self):
        pprint(
            {
                "min_bin_size": self.min_bin_size,
                "bins": self.bins,
                "packages": self.packages,
                "oversize": self.oversize,
                "bin_lookup": self.bin_lookup,
            }
        )


"""
bp = BinPack()

bp.add_bin('S', 35, 140, 200, 60)
bp.add_bin('M', 65, 170, 250, 90)
bp.add_bin('L', 75, 200, 300, 110)

bp.add_package('A', 125, 180, 11)
bp.add_package('B', 125, 180, 12)
bp.add_package('C', 148, 209, 36)
bp.add_package('D', 208, 280, 12)

# bp.debug()
pprint(bp.go())
"""



import datetime
import hashlib
import math
import pytz
import re
import time

from django.conf import settings
from django.db import connection
from django.utils.timezone import localtime


def api_auth():
    epoch = int(math.floor(time.time()))
    hash_str = "%s.%s.%s" % (
        settings.SALEBOX["API"]["KEY"],
        settings.SALEBOX["API"]["LICENSE"],
        epoch,
    )

    return {
        "pos": settings.SALEBOX["API"]["KEY"],
        "epoch": epoch,
        "hash": hashlib.sha256(hash_str.encode("utf-8")).hexdigest(),
        "software_type": "salebox_django",
        "software_version": "0.0.247",
    }


def dictfetchall(sql):
    with connection.cursor() as cursor:
        cursor.execute(sql)
        columns = [col[0] for col in cursor.description]
        return [dict(zip(columns, row)) for row in cursor.fetchall()]


def fetchflatlist(sql):
    with connection.cursor() as cursor:
        cursor.execute(sql)
        return [row[0] for row in cursor.fetchall()]


def fetchsinglevalue(sql):
    return fetchflatlist(sql)[0]


def get_rating_display(score, vote_count):
    if vote_count > 0:
        return {"100": score, "5": round(score / 20), "10": round(score / 10)}
    else:
        return {"100": 0, "5": 0, "10": 0}


def json_to_datetime_local(value):
    return localtime(json_to_datetime_utc(value))


def json_to_datetime_utc(value):
    return datetime.datetime(*value, tzinfo=pytz.UTC)


def update_natural_sort(model, str_column, sort_column):
    # retrieve all rows
    rows = list(model.objects.values("id", str_column, sort_column))

    # update string to be sortable
    for row in rows:
        row[str_column] = row[str_column].lower()
        row[str_column] = row[str_column].strip()
        row[str_column] = re.sub(r"^the\s+", "", row[str_column])
        row[str_column] = re.sub(
            r"\d+", lambda x: "%08d" % (int(x.group(0)),), row[str_column]
        )

    # sort the values
    rows = sorted(rows, key=lambda k: k[str_column])

    # update the database records that need it
    for i, row in enumerate(rows):
        if i != row[sort_column]:
            model.objects.filter(id=row["id"]).update(**{sort_column: i})



# Generated by Django 2.1.8 on 2020-12-10 03:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0017_auto_20200309_0546'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='product',
            name='inventory_flag',
        ),
        migrations.AddField(
            model_name='product',
            name='inventory_type',
            field=models.CharField(default='T', max_length=1),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='level',
            field=models.PositiveIntegerField(db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='lft',
            field=models.PositiveIntegerField(db_index=True, editable=False),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='rght',
            field=models.PositiveIntegerField(db_index=True, editable=False),
        ),
    ]



from saleboxdjango.models import Product, ProductVariant


def sync_productvariant(data):
    # get fieldname
    fieldnames = [f.name for f in ProductVariant._meta.get_fields()]

    # retrieve / create object
    o = ProductVariant.objects.filter(id=data['id']).first()
    if o is None:
        o = ProductVariant()

    # populate foreign keys
    if data['product']:
        data['product'] = Product.objects.get(id=data['product'])

    # define fields to exclude from update
    exclude = ['attribute_1', 'attribute_2', 'attribute_3', 'attribute_4', 'attribute_5', 'attribute_6', 'attribute_7', 'attribute_8', 'attribute_9', 'attribute_10']

    # save data
    for key in data:
        if key in fieldnames and key not in exclude:
            setattr(o, key, data[key])
    o.save()

    # save many to many
    for i in range(1, 11):
        key = f'attribute_{i}'
        getattr(o, key).set(data[key])



from datetime import datetime
from math import ceil

from saleboxdjango.lib.api import get
from saleboxdjango.management.commands.lib.log import log
from saleboxdjango.models import ProductVariant, KeyValueStore


class SaleboxPullInventory:
    def __init__(self):
        self.config = self.get_config()
        self.process_all()
        self.process_recent()
        self.set_config()

    def get_config(self):
        config = {"all": 0, "recent": 0}

        # get kvs from db
        kvs = KeyValueStore.objects.filter(key="sync_inventory").first()
        if kvs is not None:
            config = {**config, **kvs.value}

        return config

    def get_now(self):
        return datetime.utcnow().timestamp()

    def process_all(self):
        now = self.get_now()

        # bail if the sync has been done recently
        if (
            now - self.config["all"] < 60 * 60 * 6
        ):  # 6 hours - hardcoded now, but should be a config value
            return

        # retrieve the entire inventory
        r = get("/api/v1/inventory", {})
        if r.status_code != 200:
            log(
                "sync_inventory",
                f"API call failed, status: {r.status_code}, text: {r.text}",
                "ERROR",
            )
            return

        # update inventory database
        self.update_inventory(r.json()["inventory"])

        # set all AND recent timestamps
        self.config["all"] = now
        self.config["recent"] = now
        return

    def process_recent(self):
        now = self.get_now()

        # bail if the sync has been done recently
        diff = ceil(now - self.config["recent"])
        if diff < 30:  # 1 minute - hardcoded now, but should be a config value
            return

        # retrieve inventory changes: diff + 60 seconds
        r = get("/api/v1/inventory", {"seconds": diff + 60})
        if r.status_code != 200:
            log(
                "sync_inventory",
                f"API call failed, status: {r.status_code}, text: {r.text}",
                "ERROR",
            )
            return

        # update inventory database
        self.update_inventory(r.json()["inventory"])

        # set recent timestamp only
        self.config["recent"] = now
        return

    def set_config(self):
        kvs = KeyValueStore.objects.filter(key="sync_inventory").first()
        if kvs is None:
            KeyValueStore(key="sync_inventory", value=self.config).save()
        else:
            kvs.value = self.config
            kvs.save()

    def update_inventory(self, data):
        if len(data) == 0:
            return

        # update values in database
        lookup = {d[0]: d[1] for d in data}
        variants = ProductVariant.objects.filter(id__in=[d[0] for d in data])
        for pv in variants:
            if pv.stock_count != lookup[pv.id]:
                pv.stock_count = lookup[pv.id]
                pv.save()



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.utils.translation import gettext, gettext_lazy as _

from saleboxdjango.models import *


# inlines


class CheckoutStoreUpdateInline(admin.TabularInline):
    model = CheckoutStoreUpdate
    readonly_fields = ("status", "created", "data")
    extra = 0

    def has_delete_permission(self, request, obj=None):
        return False


class DiscountRulesetInline(admin.TabularInline):
    model = DiscountRuleset
    exclude = ["product_variant"]
    extra = 0


# models


class AnalyticAdmin(admin.ModelAdmin):
    list_display = (
        "key",
        "first_seen",
        "last_seen",
        "page_views",
        "ip_country",
        "ip_city",
        "ip_address",
        "ua_browser_family",
        "ua_os_family",
        "language",
        "screen_size",
        "ua_is_mobile",
        "ua_is_tablet",
        "ua_is_pc",
        "ua_is_bot",
    )
    list_filter = ("ua_is_bot",)
    readonly_fields = (
        "key",
        "first_seen",
        "last_seen",
        "page_views",
        "ip_address",
        "ip_country",
        "ip_city",
        "ip_tz",
        "ip_lat",
        "ip_lng",
        "ua_browser_family",
        "ua_browser_version",
        "ua_os_family",
        "ua_os_version",
        "ua_device_family",
        "ua_device_brand",
        "ua_device_model",
        "ua_is_mobile",
        "ua_is_tablet",
        "ua_is_touch_capable",
        "ua_is_pc",
        "ua_is_bot",
        "language",
        "screen_width",
        "screen_height",
    )

    def screen_size(self, obj):
        if obj.screen_height and obj.screen_width:
            return "%sx%s" % (obj.screen_width, obj.screen_height)
        else:
            return "-"


class AttributeAdmin(admin.ModelAdmin):
    list_display = ("code",)


class AttributeItemAdmin(admin.ModelAdmin):
    list_display = ("value", "slug", "attribute")
    list_filter = ("attribute",)


class BasketWishlistAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "user_or_session",
        "variant",
        "last_update",
        "quantity",
        "weight",
        "basket_flag",
    )
    list_filter = ("basket_flag",)
    search_fields = ("user__email",)

    def user_or_session(self, obj):
        if obj.user:
            return obj.user.email
        else:
            return obj.session


class CallbackStoreAdmin(admin.ModelAdmin):
    list_display = ("created", "ip_address", "method")
    search_fields = ("post",)


class CheckoutStoreAdmin(admin.ModelAdmin):
    list_display = ("created", "status", "uuid", "visible_id", "user", "gateway_code")
    inlines = [CheckoutStoreUpdateInline]
    list_filter = ("status", "gateway_code")
    readonly_fields = (
        "uuid",
        "visible_id",
        "user",
        "gateway_code",
        "status",
        "data",
        "payment_received",
    )
    search_fields = ("visible_id", "data")


class ContentKeyValueStoreAdmin(admin.ModelAdmin):
    list_display = ("name", "code", "is_locked", "is_active")


class ContentPageAdmin(admin.ModelAdmin):
    list_display = ("combined_slug", "is_locked", "is_active")


class ContentPageItemAdmin(admin.ModelAdmin):
    list_display = ("page", "item_type", "order", "is_locked", "is_active")
    list_filter = ("item_type", "is_active", "is_locked")


class CountryAdmin(admin.ModelAdmin):
    list_display = ("code", "default")


class CountryStateAdmin(admin.ModelAdmin):
    list_display = ("country", "code", "full_code")


class DiscountGroupAdmin(admin.ModelAdmin):
    inlines = [DiscountRulesetInline]
    list_display = ("name", "group_type")
    list_filter = ("group_type",)


class EventAdmin(admin.ModelAdmin):
    list_display = (
        "event",
        "salebox_member_id",
        "transaction_guid",
        "processed_flag",
        "created",
    )
    list_filter = ("event", "processed_flag")
    readonly_fields = ("event", "salebox_member_id", "transaction_guid", "created")
    search_fields = ("salebox_member_id", "transaction_guid")


class KeyValueStoreAdmin(admin.ModelAdmin):
    list_display = ("key", "value")


class LastUpdateAdmin(admin.ModelAdmin):
    list_display = ("code", "value")


class MemberGroupAdmin(admin.ModelAdmin):
    list_display = ("name", "flat_discount_percentage")


class MemberAdmin(admin.ModelAdmin):
    list_display = ("guid", "email", "name_first", "name_last", "group")
    list_filter = ("group",)
    readonly_fields = ("parent",)
    search_fields = ("guid", "email")


class ProductCategoryAdmin(admin.ModelAdmin):
    list_display = ("name", "slug_path")
    search_fields = ("name", "slug_path")


class ProductAdmin(admin.ModelAdmin):
    list_display = ("name", "category", "sold_by", "slug")
    list_filter = ("sold_by", "category")
    search_fields = ("name", "slug", "plu", "sku")


class ProductVariantAdmin(admin.ModelAdmin):
    list_display = ("product", "name", "price", "slug")
    list_filter = ("product__category",)


class ProductVariantRatingAdmin(admin.ModelAdmin):
    list_display = ("user", "variant", "rating", "created")
    search_fields = ("user__email",)


class SaleboxUserAdmin(UserAdmin):
    fieldsets = (
        (None, {"fields": ("username", "password")}),
        (_("Personal info"), {"fields": ("first_name", "last_name", "email")}),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                ),
            },
        ),
        (_("Important dates"), {"fields": ("last_login", "date_joined")}),
        ("Salebox", {"fields": ("salebox_member_id", "salebox_member_sync")}),
    )


class SyncLogAdmin(admin.ModelAdmin):
    list_display = ("status", "function_name", "message", "created")
    list_filter = ("status", "function_name")


class SyncQueueAdmin(admin.ModelAdmin):
    list_display = ("model_name", "model_id", "created")
    list_filter = ("model_name",)


class TranslationAdmin(admin.ModelAdmin):
    list_display = ("language_code", "key", "value")
    list_filter = ("language_code",)
    search_fields = ("key",)


class UserAddressAdmin(admin.ModelAdmin):
    list_display = (
        "address_group",
        "user",
        "full_name",
        "address_1",
        "address_2",
        "country_state",
        "country",
        "postcode",
    )
    list_filter = ("address_group",)
    search_fields = (
        "user__email",
        "full_name",
        "address_1",
        "address_2",
        "address_3",
        "postcode",
    )


admin.site.register(Analytic, AnalyticAdmin)
admin.site.register(Attribute, AttributeAdmin)
admin.site.register(AttributeItem, AttributeItemAdmin)
admin.site.register(BasketWishlist, BasketWishlistAdmin)
admin.site.register(CallbackStore, CallbackStoreAdmin)
admin.site.register(CheckoutStore, CheckoutStoreAdmin)
admin.site.register(ContentKeyValueStore, ContentKeyValueStoreAdmin)
admin.site.register(ContentPage, ContentPageAdmin)
admin.site.register(ContentPageItem, ContentPageItemAdmin)
admin.site.register(Country, CountryAdmin)
admin.site.register(CountryState, CountryStateAdmin)
admin.site.register(DiscountGroup, DiscountGroupAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(KeyValueStore, KeyValueStoreAdmin)
admin.site.register(LastUpdate, LastUpdateAdmin)
admin.site.register(MemberGroup, MemberGroupAdmin)
admin.site.register(Member, MemberAdmin)
admin.site.register(ProductCategory, ProductCategoryAdmin)
admin.site.register(Product, ProductAdmin)
admin.site.register(ProductVariant, ProductVariantAdmin)
admin.site.register(ProductVariantRating, ProductVariantRatingAdmin)
admin.site.register(SyncLog, SyncLogAdmin)
admin.site.register(SyncQueue, SyncQueueAdmin)
admin.site.register(Translation, TranslationAdmin)
admin.site.register(UserAddress, UserAddressAdmin)

"""

from django.contrib.auth.admin import UserAdmin
from saleboxdjango.admin import SaleboxUserAdmin

admin.site.register(SaleboxUser, SaleboxUserAdmin)

"""



import time

from django.conf import settings
from django.forms.models import model_to_dict
from django.utils.translation import get_language

from saleboxdjango.models import CheckoutStore, UserAddress


class SaleboxCheckout:
    def __init__(self, request):
        self._init_sequence()
        self._init_session(request)
        self._write_session(request)

    def get_checkout_nav(self, curr_page_name=None):
        nav = {"order": [], "lookup": {}, "previous": None, "next": None}

        for pos, s in enumerate(self.sequence["order"]):
            accessible = (
                curr_page_name is not None and self.sequence["lookup"][s]["accessible"]
            )
            current = s == curr_page_name

            # set previous / next item helpers
            if current:
                if pos > 0:
                    tmp = self.sequence["order"][pos - 1]
                    nav["previous"] = {
                        "accessible": self.sequence["lookup"][tmp]["accessible"],
                        "label": self.sequence["lookup"][tmp]["label"],
                        "page_name": tmp,
                        "path": self.sequence["lookup"][tmp]["path"],
                    }
                if pos < len(self.sequence["order"]) - 1:
                    tmp = self.sequence["order"][pos + 1]
                    nav["next"] = {
                        "accessible": self.sequence["lookup"][tmp]["accessible"],
                        "label": self.sequence["lookup"][tmp]["label"],
                        "page_name": tmp,
                        "path": self.sequence["lookup"][tmp]["path"],
                    }

            # set lookup
            nav["lookup"][s] = {
                "accessible": accessible,
                "current": current,
            }

            # set order
            nav["order"].append(
                {
                    "accessible": accessible,
                    "current": current,
                    "label": self.sequence["lookup"][s]["label"],
                    "page_name": s,
                    "path": self.sequence["lookup"][s]["path"],
                }
            )

        for pos, n in enumerate(nav["order"]):
            try:
                nav["order"][pos]["previous_accessible"] = nav["order"][pos - 1][
                    "accessible"
                ]
            except:
                nav["order"][pos]["previous_accessible"] = False
            try:
                nav["order"][pos]["next_accessible"] = nav["order"][pos + 1][
                    "accessible"
                ]
            except:
                nav["order"][pos]["next_accessible"] = False

        return nav

    def get_raw_data(self):
        return self.data

    def get_last_accessible_page(self):
        for o in reversed(self.sequence["order"]):
            if self.sequence["lookup"][o]["accessible"]:
                return self.sequence["lookup"][o]["path"]

        return None

    def get_current_page_path(self, path_name):
        return self.sequence["lookup"][path_name]["path"]

    def get_next_page(self, page_name):
        next = self.sequence["order"].index(page_name) + 1
        if next < len(self.sequence["order"]):
            return self.sequence["lookup"][self.sequence["order"][next]]["path"]
        else:
            return None

    def page_redirect(self, page_name):
        if page_name not in self.sequence["order"]:
            raise Exception("Unrecognised SaleboxCheckout page_name: %s" % page_name)

        # if basket empty, redirect to the pre-checkout page, e.g.
        # typically the shopping basket
        if len(self.data["basket"]) == 0:
            return settings.SALEBOX["CHECKOUT"]["PRE_URL"]

        # if this page is not marked accessible, i.e. the user is
        # trying to jump steps in the process, redirect them to
        # the 'last known good' page
        if not self.is_page_accessible(page_name):
            return self.get_last_accessible_page()

        return None

    def is_page_accessible(self, page_name):
        try:
            return self.sequence["lookup"][page_name]["accessible"]
        except:
            return False

    def save_to_store(
        self, user, gateway_code="default", sent_to_gateway=True, visible_id=None
    ):
        data = self.get_raw_data()
        cs = CheckoutStore(
            visible_id=visible_id,
            user=user.id if user else None,
            gateway_code=gateway_code,
            status=20 if sent_to_gateway else 10,
            data=self.get_raw_data(),
        )
        cs.save()
        return cs

    def set_basket(self, basket, request, reset_completed=True, reset_checkout=True):
        if reset_checkout:
            self._init_data()
        if reset_completed:
            self.data["completed"] = []

        # populate data object
        self.data["basket"] = basket.get_raw_data()["basket"]
        self._write_session(request)

        # return url to redirect to
        return self.sequence["lookup"][self.sequence["order"][0]]["path"]

    def set_completed(self, page_name, request):
        self.data["completed"].append(page_name)
        self._update_sequence()
        self._write_session(request)
        return self.get_next_page(page_name)

    def set_customer_email(self, email, request):
        self.data["customer"]["email"] = email
        self._write_session(request)

    def set_invoice_address(self, required, address_or_id, meta, request):
        if address_or_id is None:
            self.data["invoice_address"]["address"] = None
        else:
            if isinstance(address_or_id, int):
                self.data["invoice_address"]["address"] = model_to_dict(
                    UserAddress.objects.get(id=address_or_id)
                )
                self.data["invoice_address"]["id"] = address_or_id
            else:
                self.data["invoice_address"]["address"] = {
                    "default": address_or_id.default,
                    "address_group": address_or_id.address_group,
                    "full_name": address_or_id.full_name,
                    "address_1": address_or_id.address_1,
                    "address_2": address_or_id.address_2,
                    "address_3": address_or_id.address_3,
                    "address_4": address_or_id.address_4,
                    "address_5": address_or_id.address_5,
                    "country_state": model_to_dict(address_or_id.country_state)
                    if address_or_id.country_state
                    else None,
                    "country": model_to_dict(address_or_id.country)
                    if address_or_id.country
                    else None,
                    "id": None,
                    "postcode": address_or_id.postcode,
                    "phone_1": address_or_id.phone_1,
                    "phone_2": address_or_id.phone_2,
                    "email": address_or_id.email,
                    "string_1": address_or_id.string_1,
                    "string_2": address_or_id.string_2,
                    "tax_id": address_or_id.tax_id,
                }
                self.data["invoice_address"]["id"] = None

        self.data["invoice_address"]["meta"] = meta
        self.data["invoice_address"]["required"] = required
        self._write_session(request)

    def set_message(self, key, value, request):
        self.data["messages"] = self.data.get("messages", {})
        self.data["messages"][key] = value
        self._write_session(request)

    def set_payment_method(self, id, meta, request):
        self.data["payment_method"] = {"id": int(id), "meta": meta}
        self._write_session(request)

    def set_shipping_address(self, required, address_or_id, meta, request):
        if isinstance(address_or_id, int):
            self.data["shipping_address"]["address"] = model_to_dict(
                UserAddress.objects.get(id=address_or_id)
            )
            self.data["shipping_address"]["id"] = address_or_id
        else:
            self.data["shipping_address"]["address"] = {
                "default": address_or_id.default,
                "address_group": address_or_id.address_group,
                "full_name": address_or_id.full_name,
                "address_1": address_or_id.address_1,
                "address_2": address_or_id.address_2,
                "address_3": address_or_id.address_3,
                "address_4": address_or_id.address_4,
                "address_5": address_or_id.address_5,
                "country_state": model_to_dict(address_or_id.country_state)
                if address_or_id.country_state
                else None,
                "country": model_to_dict(address_or_id.country)
                if address_or_id.country
                else None,
                "id": None,
                "postcode": address_or_id.postcode,
                "phone_1": address_or_id.phone_1,
                "phone_2": address_or_id.phone_2,
                "email": address_or_id.email,
                "string_1": address_or_id.string_1,
                "string_2": address_or_id.string_2,
                "tax_id": address_or_id.tax_id,
            }
            self.data["shipping_address"]["id"] = None

        self.data["shipping_address"]["meta"] = meta
        self.data["shipping_address"]["required"] = required
        self._write_session(request)

    def set_shipping_method(self, id, code, label, price, meta, request):
        self.data["shipping_method"] = {
            "code": code,
            "id": id,
            "label": label,
            "price": price,
            "meta": meta,
        }
        self._write_session(request)

    def _init_data(self):
        self.data = {
            "basket": {},
            "completed": [],
            "customer": {"email": None},
            "data": {},
            "invoice_address": {
                "address": None,
                "id": None,
                "meta": None,
                "required": None,
            },
            "last_seen": int(time.time()),
            "messages": {},
            "payment_method": {
                "id": None,
                "meta": None,
            },
            "shipping_address": {
                "address": None,
                "id": None,
                "meta": None,
                "required": None,
            },
            "shipping_method": {
                "id": None,
                "label": None,
                "meta": None,
                "price": None,
            },
        }

    def _init_sequence(self):
        self.sequence = {"order": [], "lookup": {}}

        # use language url prefix? e.g. /en/checkout/address
        url_prefix = ""
        if settings.SALEBOX["CHECKOUT"].get("URL_LANGUAGE_PREFIX", False):
            url_prefix = "/%s" % get_language()

        for i, s in enumerate(settings.SALEBOX["CHECKOUT"]["SEQUENCE"]):
            self.sequence["order"].append(s[0])
            self.sequence["lookup"][s[0]] = {
                "label": s[2] if len(s) == 3 else s[0],
                "path": "%s%s" % (url_prefix, s[1]),
                "position": i,
                "complete": False,
                "accessible": i == 0,
            }

    def _init_session(self, request):
        self._init_data()

        # attempt to import data from the session
        request.session.setdefault("saleboxcheckout", None)
        if request.session["saleboxcheckout"] is not None:
            tmp = request.session["saleboxcheckout"]
            if int(time.time()) - tmp["last_seen"] < 60 * 60:  # 1 hr
                self.data = request.session["saleboxcheckout"]

        # update data
        self._update_sequence()
        self.data["last_seen"] = int(time.time())

    def _update_sequence(self):
        for i, o in enumerate(self.sequence["order"]):
            if o in self.data["completed"]:
                self.sequence["lookup"][o]["complete"] = True
                self.sequence["lookup"][o]["accessible"] = True
                try:
                    self.sequence["lookup"][self.sequence["order"][i + 1]][
                        "accessible"
                    ] = True
                except:
                    pass

    def _write_session(self, request):
        if len(self.data["basket"]) == 0:
            request.session["saleboxcheckout"] = None
        else:
            request.session["saleboxcheckout"] = self.data



from datetime import datetime
from rawquery import RawQuery

from django.conf import settings
from saleboxdjango.lib.api import post
from saleboxdjango.management.commands.lib.log import log
from saleboxdjango.models import KeyValueStore


class SaleboxPushFootfall:
    def __init__(self):
        if not settings.SALEBOX["ANALYTICS"]["SEND"]:
            return

        self.config = self.get_config()
        self.send_footfall()
        self.set_config()

    def get_config(self):
        config = {"last_update": 0}

        # get kvs from db
        kvs = KeyValueStore.objects.filter(key="sync_footfall").first()
        if kvs is not None:
            config = {**config, **kvs.value}

        return config

    def get_now(self):
        return datetime.utcnow().timestamp()

    def send_footfall(self):
        now = self.get_now()

        # bail if the sync has been done recently
        if (
            now - self.config["last_update"] < 60 * 2
        ):  # 2 minutes - hardcoded now, but should be a config value
            return

        sql = """
            SELECT          COUNT(*) AS visitor_count
                            ,DATE_PART('hour', first_seen AT TIME ZONE 'UTC')::int AS hour
                            ,DATE_PART('day', first_seen AT TIME ZONE 'UTC')::int AS day
                            ,DATE_PART('month', first_seen AT TIME ZONE 'UTC')::int AS month
                            ,DATE_PART('year', first_seen AT TIME ZONE 'UTC')::int AS year
            FROM            saleboxdjango_analytic
            WHERE           first_seen >= TIMESTAMP 'yesterday' AT TIME ZONE 'UTC'
            AND             ua_is_bot = false
            GROUP BY        hour, day, month, year;
        """

        rq = RawQuery()
        payload = {"visitors": rq.multiple_rows(sql)}
        r = post("/api/v1/footfall/ecommerce", payload)
        if r.status_code != 200:
            log(
                "sync_footfall",
                f'API call failed: "/api/v1/footfall/ecommerce" status: {r.status_code}',
                "ERROR",
            )
            self.error = True
            return

        if r.status_code == 200:
            self.config["last_update"] = now

    def set_config(self):
        kvs = KeyValueStore.objects.filter(key="sync_footfall").first()
        if kvs is None:
            KeyValueStore(key="sync_footfall", value=self.config).save()
        else:
            kvs.value = self.config
            kvs.save()



from datetime import timedelta

from django import forms
from django.contrib.gis.geoip2 import GeoIP2
from django.http import JsonResponse
from django.utils.timezone import now
from django.views.generic import View

from user_agents import parse

from saleboxdjango.models import Analytic


class SaleboxAnalyticsForm(forms.Form):
    key = forms.UUIDField()
    lang = forms.CharField(required=False)
    width = forms.IntegerField(required=False)
    height = forms.IntegerField(required=False)


class SaleboxAnalyticsView(View):
    def get_ip(self, request):
        x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
        if x_forwarded_for:
            ip = x_forwarded_for.split(",")[0]
        else:
            ip = request.META.get("REMOTE_ADDR")
        return ip

    def post(self, request):
        form = SaleboxAnalyticsForm(request.POST)
        if form.is_valid():
            row = (
                Analytic.objects.filter(key=form.cleaned_data["key"])
                .filter(first_seen__gt=now() - timedelta(hours=24))
                .first()
            )

            if row is None:
                # get ua
                ua_string = request.META.get("HTTP_USER_AGENT", None)
                if ua_string is not None:
                    ua = parse(ua_string)

                    ip_address = self.get_ip(request)
                    ip_country = None
                    ip_city = None
                    ip_tz = None
                    ip_lat = None
                    ip_lng = None
                    if ip_address is not None:
                        try:
                            g = GeoIP2()
                            tmp = g.city(ip_address)
                            ip_country = tmp["country_code"]
                            ip_city = tmp["city"]
                            ip_tz = tmp["time_zone"]
                            ip_lat = tmp["latitude"]
                            ip_lng = tmp["longitude"]
                        except:
                            pass

                    # add to db
                    row = Analytic(
                        key=form.cleaned_data["key"],
                        ip_address=ip_address,
                        ip_country=ip_country,
                        ip_city=ip_city,
                        ip_tz=ip_tz,
                        ip_lat=ip_lat,
                        ip_lng=ip_lng,
                        ua_browser_family=(
                            ua.browser.family[:32] if ua.browser.family else None
                        ),
                        ua_browser_version=(
                            ua.browser.version_string[:20]
                            if ua.browser.version_string
                            else None
                        ),
                        ua_os_family=ua.os.family[:20] if ua.os.family else None,
                        ua_os_version=(
                            ua.os.version_string[:20] if ua.os.version else None
                        ),
                        ua_device_family=(
                            ua.device.family[:32] if ua.device.family else None
                        ),
                        ua_device_brand=(
                            ua.device.brand[:20] if ua.device.brand else None
                        ),
                        ua_device_model=(
                            ua.device.model[:20] if ua.device.model else None
                        ),
                        ua_is_mobile=ua.is_mobile,
                        ua_is_tablet=ua.is_tablet,
                        ua_is_touch_capable=ua.is_touch_capable,
                        ua_is_pc=ua.is_pc,
                        ua_is_bot=ua.is_bot,
                        language=form.cleaned_data["lang"][:10],
                        screen_width=form.cleaned_data["width"],
                        screen_height=form.cleaned_data["height"],
                    )
                    row.save()
            else:
                # update page_views
                row.page_views = row.page_views + 1
                row.save()

        return JsonResponse({"thank": "you"})



from django.conf import settings
from django.utils.translation import get_language

from python2c2p.redirectapi import twoctwop_redirectapi

from saleboxdjango.lib.basket import SaleboxBasket
from saleboxdjango.views.checkout.gateway import SaleboxCheckoutGatewayView


class SaleboxProviders2C2PGatewayView(SaleboxCheckoutGatewayView):
    gateway_code = "2c2p"

    def gateway(self, request, *args, **kwargs):
        context = self.get_context_data()
        store = self.sc.save_to_store(request.user, gateway_code=self.gateway_code)

        # override start
        data = self.sc.get_raw_data()

        # get the html to POST to the gatewy
        context["html"] = self.gateway_init(data, request, store)

        # reset basket  TODO - should the basket be reset on first callback?
        basket = SaleboxBasket(request)
        basket.reset_basket(request)

        # render
        return self.render_to_response(context)

    def gateway_init(self, data, request, store):
        """
        Optionally override this method if there are different
        payment parameters you wish to send
        """

        tctp = twoctwop_redirectapi(
            settings.TWOCTWOP["MERCHANT_ID"],
            settings.TWOCTWOP["SECRET_KEY"],
            settings.TWOCTWOP["GATEWAY_URL"],
        )
        tctp.set_value("payment_option", "A")

        # total price
        total_price = data["basket"]["sale_price"] + data["shipping_method"]["price"]

        # default values
        tctp.set_value("amount", total_price)
        tctp.set_value("currency", "THB")
        tctp.set_value("customer_email", request.user.email)
        tctp.set_value("default_lang", get_language())
        tctp.set_value("order_id", store.visible_id)
        tctp.set_value("payment_description", settings.TWOCTWOP["PAYMENT_DESCRIPTION"])
        tctp.set_value(
            "result_url_1",
            settings.TWOCTWOP["CALLBACK_URL_FRONTEND"],
        )
        tctp.set_value(
            "result_url_2",
            settings.TWOCTWOP["CALLBACK_URL_BACKEND"],
        )
        tctp.set_value("user_defined_1", store.uuid)

        return tctp.request()



from saleboxdjango.lib.api import get
from saleboxdjango.management.commands.lib.log import log
from saleboxdjango.models import KeyValueStore, SyncQueue


class SaleboxPullQueue:
    def __init__(self):
        self.config = self.get_config()
        self.sync_loop()

    def sync_loop(self):
        # create api call payload
        payload = {}
        for key in self.config:
            value = self.config[key]
            payload[key] = f"{value['created']}:{value['last_update']}"

        # request from api
        r = get("/api/v1/sync", payload)
        if r.status_code != 200:
            log(
                "sync_pull_queue",
                f"API call failed, status: {r.status_code}, text: {r.text}",
                "ERROR",
            )
            return

        # process response
        response = r.json()
        for model_name in response["models"]:
            self.update_queue(
                model_name,
                response["models"][model_name]["created"],
                response["models"][model_name]["last_update"],
                response["models"][model_name]["queue"],
            )

        # loop again?
        if response["resync_now"]:
            self.sync_loop()

    def get_config(self):
        # if we need to add new models in future, just add them to
        # this list here. it'll add them to the queue automatically
        models_to_sync = [
            "content_contentkeyvaluestore",
            "content_contentpage",
            "content_contentpageitem",
            "product_productcategory",
            "product_product",
            "product_productvariant",
        ]

        # generate base config
        config = {}
        for model in models_to_sync:
            config[model] = {"created": 0, "last_update": 0}

        # get kvs from db
        kvs = KeyValueStore.objects.filter(key="sync_models_config").first()
        if kvs is not None:
            config = {**config, **kvs.value}

        return config

    def set_config(self):
        kvs = KeyValueStore.objects.filter(key="sync_models_config").first()
        if kvs is None:
            KeyValueStore(key="sync_models_config", value=self.config).save()
        else:
            kvs.value = self.config
            kvs.save()

    def update_queue(self, model_name, created, last_update, queue):
        # loop through queue
        for model_id in queue:
            exists = (
                SyncQueue.objects.filter(model_id=model_id)
                .filter(model_name=model_name)
                .count()
                > 0
            )
            if not exists:
                try:
                    SyncQueue(
                        model_id=model_id,
                        model_name=model_name,
                    ).save()
                except:
                    log(
                        "sync_pull_queue",
                        "Error inserting SyncQueue into database",
                        "ERROR",
                    )
                    return

        # if successful, update the config values
        self.config[model_name]["created"] = created
        self.config[model_name]["last_update"] = last_update
        self.set_config()



from django import forms
from django.conf import settings
from django.shortcuts import redirect

from .base import SaleboxCheckoutBaseView
from saleboxdjango.lib.basket import SaleboxBasket
from saleboxdjango.models import ProductVariant


class SaleboxCheckoutGatewayForm(forms.Form):
    pass


class SaleboxCheckoutGatewayView(SaleboxCheckoutBaseView):
    http_method_names = ["get", "post"]

    checkout_step = "gateway"
    form_class = SaleboxCheckoutGatewayForm
    template_name = "salebox/checkout/gateway.html"

    def gateway(self, request, *args, **kwargs):
        context = self.get_context_data()
        store = self.sc.save_to_store(request.user)

        # override this middle section relevant to your needs
        #
        #
        #
        context["html"] = "<form><!-- PUT YOUR CODE HERE --></form>"

        # reset the shopping basket
        basket = SaleboxBasket(request)
        basket.reset_basket(request)

        # render the gateway redirect
        return self.render_to_response(context)

    def get(self, request, *args, **kwargs):
        return self.validate(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.validate(request, *args, **kwargs)

    def validate(self, request, *args, **kwargs):
        # check everything in the basket is in stock
        if not self.validate_stock():
            # there is not enough stock available to
            # fulfil this customer's order...

            # force the basket data to be rebuilt
            request.session["saleboxbasket"]["created"] = 0

            # redirect to the basket page. in future we should add
            # some messaging to the customer here too
            return redirect(settings.SALEBOX["CHECKOUT"]["PRE_URL"])

        # check discounts (e.g. codes) are still valid
        if not self.validate_discounts():
            # something is wrong with the discount applied to this order,
            # i.e it is no longer valid somehow

            # todo
            # undo the problem here
            #
            #

            # redirect to the basket page. in future we should add
            # some messaging to the customer here too
            return redirect(settings.SALEBOX["CHECKOUT"]["PRE_URL"])

        # everything ok, proceed to payment gateway
        return self.gateway(request, *args, **kwargs)

    def validate_discounts(self):
        # todo
        #
        #

        return True

    def validate_stock(self):
        data = self.sc.get_raw_data()

        # create dict of variant_id / qty
        basket = {}
        for item in data["basket"]["items"]:
            if item["variant"]["id"] not in basket:
                basket[item["variant"]["id"]] = 0
            basket[item["variant"]["id"]] += item["qty"]

        # ignore non-inventory basket items
        non_inventory_ids = (
            ProductVariant.objects.filter(id__in=list(basket.keys()))
            .filter(product__inventory_type__in=["I", "C"])
            .values_list("id", flat=True)
        )
        for id in non_inventory_ids:
            del basket[id]

        # return False if there are more items in the basket
        # than there are available in stock
        variants = ProductVariant.objects.filter(id__in=list(basket.keys())).filter(
            preorder_flag=False
        )
        for pv in variants:
            if pv.stock_total < basket[pv.id]:
                return False

        return True



from django import forms

from saleboxdjango.views.checkout.base import SaleboxCheckoutBaseView


class SaleboxCheckoutPaymentMethodForm(forms.Form):
    payment_method = forms.ChoiceField(
        choices=(
            (1, "Credit or Debit Card"),
            (2, "E-Wallet"),
        )
    )


class SaleboxCheckoutPaymentMethodView(SaleboxCheckoutBaseView):
    checkout_step = "payment_method"
    form_class = SaleboxCheckoutPaymentMethodForm
    template_name = "salebox/checkout/payment_method.html"

    def form_valid_pre_redirect(self, form):
        id = form.cleaned_data["payment_method"]
        self.sc.set_payment_method(id, None, self.request)
        return True



from saleboxdjango.views.address.base import SaleboxAddressView
from saleboxdjango.forms import SaleboxAddressAddForm

from saleboxdjango.models import Country, CountryState, UserAddress


class SaleboxAddressAddView(SaleboxAddressView):
    action = "address-add"
    form = SaleboxAddressAddForm

    def form_valid(self, request):
        try:
            country = None
            if self.form.cleaned_data["country"] is not None:
                country = Country.objects.filter(
                    id=self.form.cleaned_data["country"]
                ).first()

            state = None
            if self.form.cleaned_data["country_state"] is not None:
                state = CountryState.objects.filter(
                    id=self.form.cleaned_data["country_state"]
                ).first()

            ua = UserAddress(
                user=request.user,
                default=self.form.cleaned_data["set_default"],
                address_group=self.form.cleaned_data["address_group"] or "default",
                full_name=self.form.cleaned_data["full_name"],
                address_1=self.form.cleaned_data["address_1"],
                address_2=self.form.cleaned_data["address_2"],
                address_3=self.form.cleaned_data["address_3"],
                address_4=self.form.cleaned_data["address_4"],
                address_5=self.form.cleaned_data["address_5"],
                country_state=state,
                country=country,
                postcode=self.form.cleaned_data["postcode"],
            )
            ua.save()
            self.status = "success"
        except:
            self.status = "fail"



# Generated by Django 3.2.6 on 2021-08-20 17:02

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0023_contentkeyvaluestore'),
    ]

    operations = [
        migrations.AddField(
            model_name='contentkeyvaluestore',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='contentkeyvaluestore',
            name='created_local',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='contentkeyvaluestore',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='contentkeyvaluestore',
            name='last_update_local',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



from opposablethumbs.opposablethumbs import OpposableThumbs

from django.conf import settings
from django.http import Http404


def salebox_image_view(request, imgtype, dir, id, hash, suffix):
    try:
        f = "%s/salebox/%s/%s.%s.%s" % (settings.MEDIA_ROOT, dir, id, hash, suffix)
        params = "file=%s&%s" % (f, settings.SALEBOX["IMG"][imgtype])
        ot = OpposableThumbs(params)
        return ot.response()
    except:
        raise Http404()



from django import forms
from django.conf import settings
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt

from saleboxdjango.models import LastUpdate


class SaleboxSyncForm(forms.Form):
    key = forms.CharField()
    license = forms.CharField()


@csrf_exempt
def SaleboxSyncView(request):
    success = False

    # is this a POST?
    if request.method == "POST":
        # did it come from the server IP?
        if settings.SALEBOX["API"]["IP"] == request.META["REMOTE_ADDR"]:
            # did it send a valid key / license combo?
            form = SaleboxSyncForm(request.POST)
            if form.is_valid():
                if settings.SALEBOX["API"]["KEY"] == str(
                    form.cleaned_data["key"]
                ) and settings.SALEBOX["API"]["LICENSE"] == str(
                    form.cleaned_data["license"]
                ):
                    # ok, we're good. reset the LastUpdate value
                    lu = (
                        LastUpdate.objects.filter(code="saleboxsync_pull_start")
                        .filter(value__gt=0)
                        .first()
                    )
                    if lu is not None:
                        lu.value = 0
                        lu.save()
                    return JsonResponse({"status": "OK"})

    # fallback
    return JsonResponse({"status": "Fail"})



# Generated by Django 3.2.6 on 2021-08-11 06:18

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0019_auto_20210807_0644'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContentPage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('page_type', models.CharField(max_length=30)),
                ('slug', models.SlugField()),
                ('combined_slug', models.CharField(max_length=255)),
                ('depth', models.IntegerField(default=0)),
                ('order', models.IntegerField(default=0)),
                ('is_locked', models.BooleanField(default=False)),
                ('is_active', models.BooleanField(default=False)),
                ('meta_data', models.JSONField(blank=True, default=dict)),
                ('parent_ids', models.CharField(blank=True, max_length=255, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.contentpage')),
            ],
        ),
        migrations.CreateModel(
            name='KeyValueStore',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(db_index=True, max_length=32, unique=True)),
                ('value', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='SyncLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('INFO', 'INFO'), ('WARNING', 'WARNING'), ('ERROR', 'ERROR')], default='INFO', max_length=7)),
                ('function_name', models.CharField(max_length=30)),
                ('message', models.CharField(max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'ordering': ('created',),
            },
        ),
        migrations.CreateModel(
            name='SyncQueue',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('model_id', models.CharField(max_length=36)),
                ('model_name', models.CharField(max_length=50)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'ordering': ('created',),
            },
        ),
        migrations.CreateModel(
            name='ContentPageItem',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('item_type', models.CharField(max_length=30)),
                ('order', models.IntegerField(default=0)),
                ('is_locked', models.BooleanField(default=False)),
                ('is_active', models.BooleanField(default=True)),
                ('content', models.JSONField(blank=True, default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('page', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.contentpage')),
            ],
        ),
    ]



# Generated by Django 3.2.6 on 2021-08-14 11:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0020_contentpage_contentpageitem_keyvaluestore_synclog_syncqueue'),
    ]

    operations = [
        migrations.RenameField(
            model_name='attribute',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='attribute',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='attributeitem',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='attributeitem',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='contentpage',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='contentpage',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='contentpageitem',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='contentpageitem',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='country',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='country',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='countrystate',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='countrystate',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='discountgroup',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='discountgroup',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='discountruleset',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='discountruleset',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='member',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='member',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='membergroup',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='membergroup',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='product',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='product',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='productcategory',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='productcategory',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='productvariant',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='productvariant',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='productvariantimage',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='productvariantimage',
            old_name='last_update',
            new_name='last_update_local',
        ),
        migrations.RenameField(
            model_name='translation',
            old_name='created',
            new_name='created_local',
        ),
        migrations.RenameField(
            model_name='translation',
            old_name='last_update',
            new_name='last_update_local',
        ),
    ]



import importlib
import time

from django.conf import settings
from django.template.loader import render_to_string

from saleboxdjango.models import BasketWishlist, ProductVariant


class SaleboxBasket:
    def __init__(self, request):
        self.cookie = None
        self.data = None

        # init a default value if the session variable is not set
        # load the basket into memory if it exists
        request.session.setdefault("saleboxbasket", None)
        if request.session["saleboxbasket"] is not None:
            self.data = request.session["saleboxbasket"]

        # if the user is not logged in, store their sessionid in a cookie
        # so we can use it later to fetch their anonymous basket
        key = request.session.session_key
        if not request.user.is_authenticated:
            if key is None:
                # just logged out, clear any lingering data
                self.data = None
            else:
                try:
                    if request.COOKIES["psessionid"] != key:
                        self.cookie = "add"
                except:
                    self.cookie = "add"

        # if the user is logged in, see if their sessionid has changed.
        # if so, they've just logged in and we need to migrate their
        # anonymous basket to their user one
        if request.user.is_authenticated:
            try:
                if request.COOKIES["psessionid"] != key:
                    self._migrate_anonymous_basket(
                        request.COOKIES["psessionid"], request.user
                    )
                    self.cookie = "remove"
                    self.data = None
            except:
                pass

        # check the time hasn't expired. if it has, rebuild the
        # basket details. this is to keep shopping baskets in sync
        # for those users using more than one device
        if self.data is not None:
            if int(time.time()) - request.session["saleboxbasket"]["created"] > 300:
                self.data = None

        # if no data exists, create it
        if self.data is None:
            self._init_basket(request)

    def debug(self):
        from pprint import pprint

        pprint(self.data)

    def get_cookie_action(self, request):
        return self.cookie

    def get_data(self, request, results, variant_id=None):
        results = results.split(",")
        o = {}

        # construct output
        if "all" in results:
            results = [
                "basket_html_button",
                "basket_html_full",
                "basket_html_summary",
                "basket_loyalty",
                "basket_price",
                "basket_qty_total",
                "basket_qty_variant",
                "wishlist_html_button",
                "wishlist_html_full",
                "wishlist_html_summary",
            ]

        if "basket_html_button" in results and variant_id:
            try:
                qty = self.data["basket"]["lookup"][str(variant_id)]["qty"]
            except:
                qty = 0

            try:
                stock_total = self.data["basket"]["lookup"][str(variant_id)]["variant"][
                    "stock_total"
                ]
            except:
                # this is a hack to prevent the button showing the
                # 'out of stock' message. Its works though - if we're
                # reaching this code, the variant is in stock
                stock_total = 1

            try:
                preorder_flag = self.data["basket"]["lookup"][str(variant_id)][
                    "variant"
                ]["preorder_flag"]
            except:
                preorder_flag = False

            o["basket_html_button"] = render_to_string(
                "salebox/basket/button_basket.html",
                {
                    "pv": {
                        "id": variant_id,
                        "basket_qty": qty,
                        "preorder_flag": preorder_flag,
                        "stock_total": stock_total,
                    },
                    "request": request,
                },
                request,
            )

        if "basket_html_full" in results:
            o["basket_html_full"] = render_to_string(
                "salebox/basket/basket_full.html",
                {"data": self.data, "request": request},
                request,
            )

        if "basket_html_summary" in results:
            o["basket_html_summary"] = render_to_string(
                "salebox/basket/basket_summary.html",
                {"data": self.data, "request": request},
                request,
            )

        if "basket_loyalty" in results:
            o["basket_loyalty"] = self.data["basket"]["loyalty"]

        if "basket_price" in results:
            o["basket_price"] = {
                "orig_price": self.data["basket"]["orig_price"],
                "sale_price": self.data["basket"]["sale_price"],
            }

        if "basket_qty_total" in results:
            o["basket_qty_total"] = self.data["basket"]["qty"]

        if "basket_qty_variant" in results and variant_id:
            o["basket_qty_variant"] = len(self.data["basket"]["order"])

        if "wishlist_html_button" in results and variant_id:
            in_wishlist = str(variant_id) in self.data["wishlist"]["lookup"]

            o["wishlist_html_button"] = render_to_string(
                "salebox/basket/button_wishlist.html",
                {
                    "pv": {"id": variant_id, "in_wishlist": in_wishlist},
                    "request": request,
                },
                request,
            )

        if "wishlist_html_full" in results:
            o["wishlist_html_full"] = render_to_string(
                "salebox/basket/wishlist_full.html",
                {"data": self.data, "request": request},
                request,
            )

        if "wishlist_html_summary" in results:
            o["wishlist_html_full"] = render_to_string(
                "salebox/basket/wishlist_summary.html",
                {"data": self.data, "request": request},
                request,
            )

        return o

    def get_raw_data(self):
        return self.data

    def migrate_basket_wishlist(self, request, variant, to_basket):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        bwl = self._filter_basket_queryset(
            request,
            BasketWishlist.objects.filter(variant=variant).exclude(
                basket_flag=to_basket
            ),
        )

        for b in bwl:
            b.basket_flag = to_basket
            b.quantity = 1
            b.save()

        # re-populate basket
        self._init_basket(request)

    def reset_basket(self, request):
        # get items from db
        self._filter_basket_queryset(
            request, BasketWishlist.objects.filter(basket_flag=True)
        ).delete()
        self._init_basket(request)

    def update_basket(self, request, variant, qty, relative, **kwargs):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        # get existing items to update
        items = self._get_variants(request, variant)

        # update db
        if items["basket"] is None:
            if items["wishlist"] is None:
                # create basket item
                if int(qty) > 0:
                    self._add_variant(request, variant, qty, True, **kwargs)
            else:
                # wishlist item exists, convert to basket
                items["wishlist"].basket_flag = True
                items["wishlist"].quantity = int(qty)
                items["wishlist"].save()
        else:
            # update basket item
            if relative:
                items["basket"].quantity += int(qty)
            else:
                items["basket"].quantity = int(qty)
            if items["basket"].quantity > 0:
                items["basket"].save()
            else:
                items["basket"].delete()

            # delete wishlist item if exists
            if items["wishlist"] is not None:
                items["wishlist"].delete()

        # re-populate basket
        self._init_basket(request)

    def update_wishlist(self, request, variant, add):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        # get existing items to update
        items = self._get_variants(request, variant)

        # update db
        if add:
            if items["wishlist"] is None:
                # add to wishlist
                self._add_variant(request, variant, 1, False)
        else:
            if items["wishlist"] is not None:
                items["wishlist"].delete()

        # re-populate basket
        self._init_basket(request)

    def _add_variant(self, request, variant, qty, basket, **kwargs):
        bwl = BasketWishlist(
            variant=variant,
            quantity=int(qty),
            weight=int(kwargs['weight']) if 'weight' in kwargs else 0,
            basket_flag=basket,
        )

        if request.user.is_authenticated:
            bwl.user = request.user
        else:
            bwl.session = request.session.session_key

        bwl.save()

    def _calculate_loyalty(self, request):
        # pre_calculate_loyalty
        self._call_external("PRE_CALCULATE_LOYALTY", request)

        for i in self.data["basket"]["items"]:
            if i["variant"]["loyalty"] is not None:
                i["variant"]["qty_loyalty"] = i["variant"]["loyalty"] * i["qty"]
                self.data["basket"]["loyalty"] += i["variant"]["loyalty"] * i["qty"]

        # post_calculate_loyalty
        self._call_external("POST_CALCULATE_LOYALTY", request)

    def _calculate_price(self, request):
        # pre_calculate_price
        self._call_external("PRE_CALCULATE_PRICE", request)

        for i in self.data["basket"]["items"]:
            if i.get('weight', 0) == 0:
                weight_modifier = 1
            else:
                weight_modifier = i['weight']

            # price for this row item
            i["variant"]["qty_price"] = i["variant"]["price"] * i["qty"] * weight_modifier
            i["variant"]["qty_sale_price"] = i["variant"]["sale_price"] * i["qty"] * weight_modifier

            # total basket price
            self.data["basket"]["orig_price"] += i["variant"]["qty_price"]
            self.data["basket"]["sale_price"] += i["variant"]["qty_sale_price"]
            if i["variant"]["preorder_flag"]:
                self.data["basket"]["orig_price_preorder"] += i["variant"]["qty_price"]
                self.data["basket"]["sale_price_preorder"] += i["variant"]["qty_sale_price"]
            else:
                self.data["basket"]["orig_price_onshelf"] += i["variant"]["qty_price"]
                self.data["basket"]["sale_price_onshelf"] += i["variant"]["qty_sale_price"]

        self.data["basket"]["saving"] = (
            self.data["basket"]["orig_price"] - self.data["basket"]["sale_price"]
        )

        # post_calculate_price
        self._call_external("POST_CALCULATE_PRICE", request)

    def _call_external(self, name, request):
        if settings.SALEBOX["CHECKOUT"]:
            funcstr = settings.SALEBOX["CHECKOUT"].get(name)
            if funcstr:
                pkg, attr = funcstr.rsplit(".", 1)
                func = getattr(importlib.import_module(pkg), attr)
                self.data = func(self.data, request)

    def _filter_basket_queryset(self, request, qs):
        qs = qs.select_related(
            "variant",
            "variant__product",
            "variant__product__category",
        )
        qs = qs.order_by("created")

        if request.user.is_authenticated:
            return qs.filter(user=request.user).filter(session__isnull=True)
        else:
            return qs.filter(user__isnull=True).filter(
                session=request.session.session_key
            )

    def _get_raw_basket(self, request):
        while True:
            # get basket items
            basket = self._filter_basket_queryset(
                request, BasketWishlist.objects.filter(basket_flag=True)
            )

            # do not allow basket qty to exceed available qty
            for b in basket:
                if (
                    not b.variant.preorder_flag
                    and b.variant.product.inventory_type == "T"
                ):
                    if b.quantity > 0 and b.quantity > b.variant.stock_total:
                        if b.variant.stock_total < 1:
                            b.quantity = 1
                            b.basket_flag = False
                        else:
                            b.quantity = b.variant.stock_total

                        b.save()

            # get items from db
            qs = self._filter_basket_queryset(request, BasketWishlist.objects.all())

            # init duplicate lookups
            duplicate_basket = {}
            duplicate_wishlist = {}
            delete_ids = []

            # find duplicates
            for pos, q in enumerate(qs):
                vid = str(q.variant.id)

                if q.basket_flag:
                    # basket
                    if vid in duplicate_basket:
                        # merge into the first instance of this variant
                        qs[duplicate_basket[vid]].quantity += q.quantity
                        qs[duplicate_basket[vid]].save()

                        # mark the duplicate to be deleted
                        delete_ids.append(q.id)
                    else:
                        duplicate_basket[vid] = pos
                else:
                    # wishlist
                    if vid in duplicate_wishlist:
                        delete_ids.append(q.id)
                    else:
                        if q.quantity != 1:
                            q.quantity = 1
                            q.save()
                        duplicate_wishlist[vid] = pos

            # delete duplicates if they exist
            if len(delete_ids) > 0:
                BasketWishlist.objects.filter(id__in=delete_ids).delete()
            else:
                break

        return qs

    def _get_variants(self, request, variant):
        if isinstance(variant, int):
            variant = ProductVariant.objects.get(id=variant)

        basket = self._filter_basket_queryset(
            request,
            BasketWishlist.objects.filter(basket_flag=True).filter(variant=variant),
        )

        wishlist = self._filter_basket_queryset(
            request,
            BasketWishlist.objects.filter(basket_flag=False).filter(variant=variant),
        )

        return {
            "basket": basket.first(),
            "wishlist": wishlist.first(),
        }

    def _init_basket_attribute(self, attribute_item):
        items = attribute_item.all()
        output = []

        for item in items:
            output.append(
                {
                    "attribute": item.attribute.code if item.attribute else None,
                    "id": item.id,
                    "slug": item.slug,
                    "value": item.value,
                }
            )

        return output

    def _init_basket(self, request):
        self.data = {
            "created": int(time.time()),
            "basket": {
                "qty": 0,
                "qty_onshelf": 0,
                "qty_preorder": 0,
                "orig_price": 0,
                "orig_price_onshelf": 0,
                "orig_price_preorder": 0,
                "sale_price": 0,
                "sale_price_onshelf": 0,
                "sale_price_preorder": 0,
                "loyalty": 0,
                "items": [],
                "order": [],
                "lookup": {},
            },
            "wishlist": {
                "qty": 0,
                "qty_onshelf": 0,
                "qty_preorder": 0,
                "order": [],
                "items": [],
                "lookup": {},
            },
        }

        # retrieve items from db
        qs = self._get_raw_basket(request)

        # populate data
        for q in qs:
            pv = {
                "attribute_1": self._init_basket_attribute(q.variant.attribute_1),
                "attribute_2": self._init_basket_attribute(q.variant.attribute_2),
                "attribute_3": self._init_basket_attribute(q.variant.attribute_3),
                "attribute_4": self._init_basket_attribute(q.variant.attribute_4),
                "attribute_5": self._init_basket_attribute(q.variant.attribute_5),
                "attribute_6": self._init_basket_attribute(q.variant.attribute_6),
                "attribute_7": self._init_basket_attribute(q.variant.attribute_7),
                "attribute_8": self._init_basket_attribute(q.variant.attribute_8),
                "attribute_9": self._init_basket_attribute(q.variant.attribute_9),
                "attribute_10": self._init_basket_attribute(q.variant.attribute_10),
                "boolean_1": q.variant.boolean_1,
                "boolean_2": q.variant.boolean_2,
                "boolean_3": q.variant.boolean_3,
                "boolean_4": q.variant.boolean_4,
                "boolean_5": q.variant.boolean_5,
                "date_1": q.variant.date_1,
                "date_2": q.variant.date_2,
                "default_image": q.variant.default_image,
                "ecommerce_description": q.variant.ecommerce_description,
                "id": q.variant.id,
                "image": q.variant.image,
                "int_1": q.variant.int_1,
                "int_2": q.variant.int_2,
                "int_3": q.variant.int_3,
                "int_4": q.variant.int_4,
                "int_5": q.variant.int_5,
                "local_image": q.variant.local_image,
                "loyalty": q.variant.loyalty_points,
                "member_discount_applicable": q.variant.member_discount_applicable,
                "name": q.variant.name,
                "plu": q.variant.plu,
                "preorder_flag": q.variant.preorder_flag,
                "price": q.variant.price,
                "product": {
                    "attribute_1": self._init_basket_attribute(
                        q.variant.product.attribute_1
                    ),
                    "attribute_2": self._init_basket_attribute(
                        q.variant.product.attribute_2
                    ),
                    "attribute_3": self._init_basket_attribute(
                        q.variant.product.attribute_3
                    ),
                    "attribute_4": self._init_basket_attribute(
                        q.variant.product.attribute_4
                    ),
                    "attribute_5": self._init_basket_attribute(
                        q.variant.product.attribute_5
                    ),
                    "attribute_6": self._init_basket_attribute(
                        q.variant.product.attribute_6
                    ),
                    "attribute_7": self._init_basket_attribute(
                        q.variant.product.attribute_7
                    ),
                    "attribute_8": self._init_basket_attribute(
                        q.variant.product.attribute_8
                    ),
                    "attribute_9": self._init_basket_attribute(
                        q.variant.product.attribute_9
                    ),
                    "attribute_10": self._init_basket_attribute(
                        q.variant.product.attribute_10
                    ),
                    "category": {
                        "id": q.variant.product.category.id,
                        "name": q.variant.product.category.name,
                        "short_name": q.variant.product.category.short_name,
                    },
                    "id": q.variant.product.id,
                    "image": q.variant.product.image,
                    "local_image": q.variant.product.local_image,
                    "name": q.variant.product.name,
                    "slug": q.variant.product.slug,
                    "sold_by": q.variant.product.sold_by,
                    "string_1": q.variant.product.string_1,
                    "string_2": q.variant.product.string_2,
                    "string_3": q.variant.product.string_3,
                    "string_4": q.variant.product.string_4,
                    "vat_applicable": q.variant.product.vat_applicable,
                },
                "rating_score": q.variant.rating_score,
                "rating_vote_count": q.variant.rating_vote_count,
                "sale_percent": q.variant.sale_percent,
                "sale_price": q.variant.sale_price,
                "shipping_weight": q.variant.shipping_weight,
                "size_uom": q.variant.size_uom,
                "sku": q.variant.sku,
                "slug": q.variant.slug,
                "stock_total": q.variant.stock_total,
                "string_1": q.variant.string_1,
                "string_2": q.variant.string_2,
                "string_3": q.variant.string_3,
                "string_4": q.variant.string_4,
                "string_5": q.variant.string_5,
            }

            if q.basket_flag:
                self.data["basket"]["qty"] += q.quantity
                if q.variant.preorder_flag:
                    self.data["basket"]["qty_preorder"] += q.quantity
                else:
                    self.data["basket"]["qty_onshelf"] += q.quantity
                self.data["basket"]["order"].append(str(q.variant.id))
                self.data["basket"]["lookup"][str(q.variant.id)] = {
                    "qty": q.quantity,
                    "weight": q.weight,
                    "variant": pv,
                }
            else:
                self.data["wishlist"]["qty"] += q.quantity
                if q.variant.preorder_flag:
                    self.data["wishlist"]["qty_preorder"] += q.quantity
                else:
                    self.data["wishlist"]["qty_onshelf"] += q.quantity
                self.data["wishlist"]["order"].append(str(q.variant.id))
                self.data["wishlist"]["lookup"][str(q.variant.id)] = pv

        # populate the items list
        for i in self.data["basket"]["order"]:
            self.data["basket"]["items"].append(self.data["basket"]["lookup"][str(i)])
        for i in self.data["wishlist"]["order"]:
            self.data["wishlist"]["items"].append(
                self.data["wishlist"]["lookup"][str(i)]
            )

        # calculate basket value + loyalty points
        self._calculate_price(request)
        self._calculate_loyalty(request)

        # save to session
        request.session["saleboxbasket"] = self.data

    def _migrate_anonymous_basket(self, key, user):
        # get all basket items from previously anonymous visitor
        basket = BasketWishlist.objects.filter(user__isnull=True).filter(session=key)
        for b in basket:
            b.user = user
            b.session = None
            b.save()



from saleboxdjango.views.address.base import SaleboxAddressView
from saleboxdjango.forms import SaleboxAddressIDForm


class SaleboxAddressRemoveView(SaleboxAddressView):
    action = "address-remove"
    form = SaleboxAddressIDForm

    def form_valid(self, request):
        try:
            self.sa.set_inactive(self.form.cleaned_data["id"])
            self.status = "success"
        except:
            self.status = "fail"



# Generated by Django 2.1.8 on 2019-12-16 05:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0012_auto_20191113_1426'),
    ]

    operations = [
        migrations.CreateModel(
            name='Analytic',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.UUIDField(db_index=True, editable=False)),
                ('first_seen', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('last_seen', models.DateTimeField(auto_now=True)),
                ('page_views', models.IntegerField(default=1)),
                ('ip_address', models.GenericIPAddressField(blank=True, null=True)),
                ('ua_browser_family', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_browser_version', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_os_family', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_os_version', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_device_family', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_device_brand', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_device_model', models.CharField(blank=True, max_length=20, null=True)),
                ('ua_is_mobile', models.BooleanField(null=True)),
                ('ua_is_tablet', models.BooleanField(null=True)),
                ('ua_is_touch_capable', models.BooleanField(null=True)),
                ('ua_is_pc', models.BooleanField(null=True)),
                ('ua_is_bot', models.BooleanField(null=True)),
            ],
        ),
    ]



from django.core.mail import mail_admins
from saleboxdjango.models import SyncLog


def log(function_name, message, status="INFO"):
    SyncLog(status=status, function_name=function_name, message=message[:255]).save()

    # admin emails
    if status == "ERROR":
        mail_admins("saleboxsync2 error", message)
